/*
 * Copyright (c) 2009 Stphane Thibaudeau
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Stphane Thibaudeau - initial implementation
 */

package fr.obeo.acceleo.gen.ui.editors.template.actions;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.IEditorActionDelegate;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.FileEditorInput;

import fr.obeo.acceleo.gen.template.TemplateSyntaxExceptions;
import fr.obeo.acceleo.gen.ui.AcceleoEcoreGenUiPlugin;
import fr.obeo.acceleo.gen.ui.AcceleoGenUIMessages;
import fr.obeo.acceleo.gen.ui.editors.reflective.AcceleoReflectiveEditor;
import fr.obeo.acceleo.gen.ui.editors.template.AcceleoTemplateEditor;
import fr.obeo.acceleo.tools.ui.resources.FileSelectionDialog;

/**
 * Select a model file used to test the current template.
 * 
 * @author stephane.thibaudeau@gmail.com
 * 
 */
public class AcceleoTemplateTestOnModelAction implements IEditorActionDelegate {

	/** Title used for the file selection dialog */
	private static final String MODEL_SELECTION_TITLE = AcceleoGenUIMessages.getString("AcceleoTemplateTestOnModelAction.ModelSelectionTitle"); //$NON-NLS-1$
	/** Acceleo Reflective Editor ID */
	private static final String ACCELEO_REFLECTIVE_EDITOR_ID = "fr.obeo.acceleo.gen.ui.editors.reflective.AcceleoReflectiveEditor"; //$NON-NLS-1$
	/** Error message when no reflective editor could be opened */
	private static final String COULD_NOT_OPEN_REFLECTIVE_EDITOR = "AcceleoTemplateTestOnModelAction.CouldNotOpenReflectiveEditor"; //$NON-NLS-1$
	
	/**
	 * The template editor.
	 */
	private AcceleoTemplateEditor templateEditor = null;

	
	/**
	 * Constructor
	 */
	public AcceleoTemplateTestOnModelAction() {
	}
	
	/* (non-Javadoc) */
	public void setActiveEditor(IAction action, IEditorPart targetEditor) {
		// Sets the active editor only if it's an acceleo template editor
		if (targetEditor instanceof AcceleoTemplateEditor) {
			this.templateEditor = (AcceleoTemplateEditor)targetEditor;
		} else {
			this.templateEditor = null;
		}
	}
	
	/* (non-Javadoc) */
	public void run(IAction action) {
		// Do nothing if the action has no associated template editor
		if (templateEditor == null) {
			return;
		}			
		
		// Creates a dialog to select the model file
		FileSelectionDialog dialog = new FileSelectionDialog(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
				MODEL_SELECTION_TITLE, 1, new String[]{"*"}, false); //$NON-NLS-1$

		// Preselects the last model used for testing
		IFile lastModel = templateEditor.getLastTestingModel();
		if (lastModel != null && lastModel.isAccessible()) {
			dialog.setInitialSelections(new Object[]{lastModel.getFullPath()});
		}
		
		// Opens the dialog and gets the result
		dialog.open();
		Object[] result = dialog.getResult();

		if (result != null && result.length > 0 && result[0] instanceof IPath) {
			IFile modelFile = ResourcesPlugin.getWorkspace().getRoot().getFile((IPath) result[0]);
			if (modelFile != null && modelFile.exists()) {
				// Opens an Acceleo reflective editor on the model
				AcceleoReflectiveEditor reflectiveEditor = null;
				IWorkbenchPage page = templateEditor.getEditorSite().getPage();
				try {
					// We want an editor on the given file and with the given type
					// The flags are there so that a new Acceleo Reflective Editor is opened
					// when there is already one active editor on the model but not the good kind
					reflectiveEditor = (AcceleoReflectiveEditor)page.openEditor(	new FileEditorInput(modelFile),
																			ACCELEO_REFLECTIVE_EDITOR_ID,
																			true,
																			IWorkbenchPage.MATCH_INPUT | IWorkbenchPage.MATCH_ID);
				} catch (PartInitException e) {
					AcceleoEcoreGenUiPlugin.getDefault().log(AcceleoGenUIMessages.getString(COULD_NOT_OPEN_REFLECTIVE_EDITOR, new Object[]{modelFile.getFullPath().toString()}), true);
				}
				
				// Tests if the reflective editor could be opened
				if (reflectiveEditor == null) {
					AcceleoEcoreGenUiPlugin.getDefault().log(AcceleoGenUIMessages.getString(COULD_NOT_OPEN_REFLECTIVE_EDITOR, new Object[]{modelFile.getFullPath().toString()}), true);
				}
				
				// Sets the selected model file as the last model used to test the generator
				templateEditor.setLastTestingModel(modelFile);
				
				// Associates the current generator to the model				
				if (reflectiveEditor.getActivePreviewEditor() != null) {
					try {
						reflectiveEditor.getActivePreviewEditor().selectGenerator(templateEditor.getGenerator().getFile(), null);
					} catch (TemplateSyntaxExceptions e) {
						AcceleoEcoreGenUiPlugin.getDefault().log(e, true);
					} catch (CoreException e) {
						AcceleoEcoreGenUiPlugin.getDefault().log(e, true);
					}
				} else {
					AcceleoEcoreGenUiPlugin.getDefault().log(AcceleoReflectiveEditor.ACTIVE_PREVIEW_NOT_FOUND_MESSAGE, true);
				}				

				// Activates the "source" page
				reflectiveEditor.activateSourceEditorPage();
			}
		}
	}

	/* (non-Javadoc) */
	public void selectionChanged(IAction action, ISelection selection) {
	}

}
