/*
 * Copyright (c) 2005-2010 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.ui.wizards.product;

import fr.obeo.acceleo.gen.ui.AcceleoEcoreGenUiPlugin;
import fr.obeo.acceleo.gen.ui.AcceleoGenUIMessages;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.pde.internal.ui.PDEPluginImages;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.ui.IExportWizard;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.dialogs.WizardNewProjectCreationPage;
import org.eclipse.ui.dialogs.WizardNewProjectReferencePage;

/**
 * A wizard to create a new product (i.e an eclipse plug-in) for one or several
 * Acceleo modules.
 * 
 * @author <a href="mailto:jonathan.musset@obeo.fr">Jonathan Musset</a>
 */
public class AcceleoExportToProductWizard extends Wizard implements IExportWizard {

	protected String productFileName;

	/**
	 * The current selection.
	 */
	protected IStructuredSelection selection;

	/**
	 * This is a new product wizard page.
	 */
	private WizardNewProjectCreationPage newProductPage;

	/**
	 * This is a wizard page to select the Acceleo modules to be considered.
	 */
	private WizardNewProjectReferencePage projectReferencesPage;

	/**
	 * This is a wizard page to export the new product.
	 */
	protected ProductExportWizardPage productExportWizardPage;

	/**
	 * Constructor.
	 */
	public AcceleoExportToProductWizard() {
		super();
		setDefaultPageImageDescriptor(PDEPluginImages.DESC_PRODUCT_EXPORT_WIZ);
		setWindowTitle(AcceleoGenUIMessages.getString("AcceleoExportToProductWizard.Title")); //$NON-NLS-1$
		productFileName = "acceleoModule.product"; //$NON-NLS-1$
	}

	/**
	 * Gets the new product wizard page.
	 * 
	 * @return the new product wizard page
	 */
	public WizardNewProjectCreationPage getNewProductPage() {
		return newProductPage;
	}

	/**
	 * Gets the wizard page where the Acceleo modules have been selected.
	 * 
	 * @return the Acceleo modules wizard page
	 */
	public WizardNewProjectReferencePage getProjectReferencesPage() {
		return projectReferencesPage;
	}

	/* (non-Javadoc) */
	public void init(IWorkbench workbench, IStructuredSelection selection) {
		this.selection = selection;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.eclipse.jface.wizard.Wizard#addPages()
	 */
	public void addPages() {
		IProject selectedProject = getSelectedProject();
		ImageDescriptor wizardImage = AcceleoEcoreGenUiPlugin.getImageDescriptor("icons/wizard/product.gif"); //$NON-NLS-1$
		newProductPage = new WizardNewProjectCreationPage(AcceleoGenUIMessages.getString("AcceleoExportToProductWizard.NewAcceleoProduct")); //$NON-NLS-1$
		if (selectedProject != null) {
			newProductPage.setInitialProjectName(selectedProject.getName() + ".product"); //$NON-NLS-1$
		} else {
			newProductPage.setInitialProjectName("org.acceleo.module.???.product"); //$NON-NLS-1$
		}
		newProductPage.setTitle(AcceleoGenUIMessages.getString("AcceleoExportToProductWizard.CreateNewProduct")); //$NON-NLS-1$
		newProductPage.setDescription(AcceleoGenUIMessages.getString("AcceleoExportToProductWizard.TypeProductName")); //$NON-NLS-1$
		newProductPage.setImageDescriptor(wizardImage);
		addPage(newProductPage);
		projectReferencesPage = new WizardNewProjectReferencePage(AcceleoGenUIMessages.getString("AcceleoExportToProductWizard.ModuleSelection")); //$NON-NLS-1$
		projectReferencesPage.setDescription(AcceleoGenUIMessages.getString("AcceleoExportToProductWizard.SelectReferencedModules")); //$NON-NLS-1$
		projectReferencesPage.setImageDescriptor(wizardImage);
		addPage(projectReferencesPage);
		productExportWizardPage = new ProductExportWizardPage(selection);
		addPage(productExportWizardPage);
		super.addPages();
	}

	/**
	 * Gets the current selection in the workbench, and returns the
	 * corresponding project.
	 * 
	 * @return the selected project
	 */
	protected IProject getSelectedProject() {
		IProject selectedProject = null;
		if (selection != null && !selection.isEmpty()) {
			if (selection.size() > 0) {
				Object element = selection.getFirstElement();
				if (element instanceof IAdaptable) {
					element = ((IAdaptable) element).getAdapter(IResource.class);
				}
				if (element instanceof IProject) {
					selectedProject = (IProject) element;
				} else if (element instanceof IResource) {
					selectedProject = ((IResource) element).getProject();
				}
			}
		}
		return selectedProject;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.eclipse.jface.wizard.Wizard#createPageControls(org.eclipse.swt.widgets.Composite)
	 */
	public void createPageControls(Composite pageContainer) {
		super.createPageControls(pageContainer);
		IProject selectedProject = getSelectedProject();
		if (selectedProject != null && ((Composite) projectReferencesPage.getControl()).getChildren().length > 1 && ((Composite) projectReferencesPage.getControl()).getChildren()[1] instanceof Table) {
			Table table = (Table) ((Composite) projectReferencesPage.getControl()).getChildren()[1];
			TableItem[] children = table.getItems();
			for (int i = 0; i < children.length; i++) {
				TableItem item = children[i];
				Object data = item.getData();
				if (data instanceof IProject && selectedProject.getName().equals(((IProject) data).getName())) {
					item.setChecked(true);
				}
			}
		}
	}

	/**
	 * Initializes the plug-ins dependencies and the API which runs the code
	 * generation by visiting each project reference.
	 * 
	 * @param page
	 *            is the standard project reference page
	 */
	protected List computeTemplatesConfiguration(WizardNewProjectReferencePage page) {
		List res = new ArrayList();
		if (page.getReferencedProjects() != null) {
			IProject[] projects = page.getReferencedProjects();
			for (int i = 0; i < projects.length; i++) {
				IProject project = projects[i];
				res.add(project.getName());
			}
		}
		return res;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.eclipse.jface.wizard.Wizard#performFinish()
	 */
	public boolean performFinish() {
		ProductData productData = createProductData();
		productData.setDestinationPath(new Path(productExportWizardPage.getDestination()));
		productData.setLocationPath(getNewProductPage().getLocationPath());
		productData.setProductFileName(productFileName);
		productData.setProjectName(getNewProductPage().getProjectName());
		productData.setSelectedProject(getSelectedProject());
		productData.setWizardContainer(getContainer());

		IRunnableWithProgress create = createProductRunnable(productData);
		try {
			PlatformUI.getWorkbench().getProgressService().run(false, false, create);
			return true;
		} catch (InvocationTargetException e) {
			IStatus status = new Status(IStatus.ERROR, AcceleoEcoreGenUiPlugin.getDefault().getID(), IStatus.OK, e.getMessage(), e);
			AcceleoEcoreGenUiPlugin.getDefault().getLog().log(status);
			return false;
		} catch (InterruptedException e) {
			IStatus status = new Status(IStatus.ERROR, AcceleoEcoreGenUiPlugin.getDefault().getID(), IStatus.OK, e.getMessage(), e);
			AcceleoEcoreGenUiPlugin.getDefault().getLog().log(status);
			return false;
		}
	}

	protected ProductData createProductData() {
		ProductData res = new AcceleoCreateModuleProductData(getNewProductPage().getProjectName(), computeTemplatesConfiguration(getProjectReferencesPage()));
		return res;
	}

	protected IRunnableWithProgress createProductRunnable(ProductData productData) {
		return new CreateAcceleoProductRunnable((AcceleoCreateModuleProductData) productData);
	}
}
