/*
 * Copyright (c) 2005-2010 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package fr.obeo.acceleo.template.extender.api;

import fr.obeo.acceleo.ecore.factories.FactoryException;
import fr.obeo.acceleo.gen.template.Template;
import fr.obeo.acceleo.gen.template.TemplateSyntaxExceptions;
import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeException;
import fr.obeo.acceleo.gen.template.eval.LaunchManager;
import fr.obeo.acceleo.gen.template.scripts.IScript;
import fr.obeo.acceleo.gen.template.scripts.SpecificScript;
import fr.obeo.acceleo.template.extender.AcceleoExtenderMessages;
import fr.obeo.acceleo.template.extender.AcceleoExtenderPlugin;
import fr.obeo.acceleo.tools.plugins.AcceleoModuleProvider;

import java.io.File;

import org.eclipse.core.runtime.Path;
import org.eclipse.emf.ecore.EObject;

/**
 * Allows to launch a Acceleo generation using extender mechanism.
 * 
 * @author Guillaume Gebhart <a
 *         href="mailto:guillaume.gebhart@obeo.fr">guillaume.gebhart@obeo.fr</a>
 * 
 */
public class AcceleoExtenderGenerator {

	private IScript aScript;

	private String templateID;

	/**
	 * ACCELEO RUN MODE
	 */
	public static String RUN_MODE = "run";

	/**
	 * ACCELEO DEBUG MODE
	 */
	public static String DEBUG_MODE = "debug";

	/**
	 * Constructor
	 * 
	 * @param templateID
	 */
	public AcceleoExtenderGenerator(String templateID) {

		if (!(templateID == null)) {
			setTemplate(templateID);
		}
	}

	/**
	 * Initializes the template
	 * 
	 * @param templateID
	 *            the name of the template
	 * @throws Exception
	 */
	private void initializeTemplate(String templateID) throws Exception {
		File script = AcceleoModuleProvider.getDefault().getFile(new Path(templateID));
		if (script == null || !script.exists()) {
			throw new Exception(AcceleoExtenderMessages.getString("ChainRunner.Task.LaunchingChain", new Object[] { templateID, }));
		} else {
			try {
				aScript = new SpecificScript(script);
				aScript.reset();
				this.templateID = templateID;
			} catch (TemplateSyntaxExceptions e) {
				AcceleoExtenderPlugin.getDefault().log(e.getMessage(), true);
			}
		}
	}

	/**
	 * Allows to generate
	 * 
	 * @param eObject
	 *            used for the transformation
	 * @param mode
	 *            "run" or "debug"
	 * @return
	 */
	public ENode generate(final EObject eObject, final String mode) {
		ENode result = null;
		aScript = AcceleoExtender.run(aScript);

		try {
			result = evaluate(eObject, LaunchManager.create(mode, true));
		} catch (FactoryException e) {
			AcceleoExtenderPlugin.getDefault().log(e.getMessage(), true);
		} catch (ENodeException e) {
			AcceleoExtenderPlugin.getDefault().log(e.getMessage(), true);
		}

		return result;
	}

	protected ENode evaluate(final EObject object, final LaunchManager mode) throws FactoryException, ENodeException {
		Template template = aScript.getRootTemplate(object, true);

		if (template != null) {
			boolean withComment = aScript.isDefault() || !aScript.hasFileTemplate();
			if (withComment) {
				return template.evaluateWithComment(object, mode);
			} else {
				return template.evaluate(object, mode);
			}
		} else {
			return new ENode(ENode.EMPTY, object, Template.EMPTY, mode.isSynchronize());
		}
	}

	public void setTemplate(String templateID) {

		if (!templateID.equals(this.templateID)) {
			try {
				initializeTemplate(templateID);

			} catch (Exception e) {
				AcceleoExtenderPlugin.getDefault().log(e.getMessage(), true);
			}
		}
	}

	/**
	 * get the template file path
	 */
	public String getTemplateId() {
		return templateID;
	}

	/**
	 * get the template
	 */
	public IScript getTemplate() {
		return aScript;
	}

}
